@@ -1,16 +1,27 @@
 #!/usr/bin/env python3
+"""Capital Gain Calculator main module."""
+from __future__ import annotations
 
 from collections import defaultdict
 import datetime
 import decimal
 from decimal import Decimal
+import importlib.metadata
+import logging
+from pathlib import Path
 import sys
-from typing import Dict, List, Tuple
 
 from . import render_latex
 from .args_parser import create_parser
+from .const import BED_AND_BREAKFAST_DAYS, CAPITAL_GAIN_ALLOWANCES, INTERNAL_START_DATE
 from .currency_converter import CurrencyConverter
-from .dates import date_from_index, date_to_index, internal_start_date, is_date
+from .dates import (
+    date_from_index,
+    date_to_index,
+    get_tax_year_end,
+    get_tax_year_start,
+    is_date,
+)
 from .exceptions import (
     AmountMissingError,
     CalculatedAmountDiscrepancy,
@@ -21,13 +32,13 @@ from .exceptions import (
     SymbolMissingError,
 )
 from .initial_prices import InitialPrices
-from .misc import round_decimal
 from .model import (
     ActionType,
     BrokerTransaction,
     CalculationEntry,
     CalculationLog,
     CapitalGainsReport,
+    HmrcTransactionLog,
     RuleType,
 )
 from .parsers import (
@@ -35,74 +46,39 @@ from .parsers import (
     read_gbp_prices_history,
     read_initial_prices,
 )
+from .transaction_log import add_to_list, has_key
+from .util import round_decimal
 
-# For mapping of dates to int
-HmrcTransactionLog = Dict[int, Dict[str, Tuple[Decimal, Decimal, Decimal]]]
-
-
-def has_key(transactions: HmrcTransactionLog, date_index: int, symbol: str) -> bool:
-    return date_index in transactions and symbol in transactions[date_index]
+LOGGER = logging.getLogger(__name__)
 
 
 class CapitalGainsCalculator:
-    # Allowances
-    # https://www.gov.uk/guidance/capital-gains-tax-rates-and-allowances#tax-free-allowances-for-capital-gains-tax
-    CAPITAL_GAIN_ALLOWANCES: Dict[int, int] = {
-        2014: 11000,
-        2015: 11100,
-        2016: 11100,
-        2017: 11300,
-        2018: 11700,
-        2019: 12000,
-        2020: 12300,
-    }
+    """Main calculator class."""
 
     def __init__(
         self, tax_year: int, converter: CurrencyConverter, initial_prices: InitialPrices
     ):
+        """Create calculator object."""
         self.tax_year = tax_year
 
-        # 6 April
-        self.tax_year_start_date = datetime.date(tax_year, 4, 6)
-        # 5 April
-        self.tax_year_end_date = datetime.date(tax_year + 1, 4, 5)
+        self.tax_year_start_date = get_tax_year_start(tax_year)
+        self.tax_year_end_date = get_tax_year_end(tax_year)
 
         self.converter = converter
         self.initial_prices = initial_prices
 
     def date_in_tax_year(self, date: datetime.date) -> bool:
+        """Check if date is within current tax year."""
         assert is_date(date)
-        return self.tax_year_start_date <= date and date <= self.tax_year_end_date
-
-    @staticmethod
-    def add_to_list(
-        current_list: HmrcTransactionLog,
-        date_index: int,
-        symbol: str,
-        quantity: Decimal,
-        amount: Decimal,
-        fees: Decimal,
-    ) -> None:
-        # assert quantity is not None
-        if date_index not in current_list:
-            current_list[date_index] = {}
-        if symbol not in current_list[date_index]:
-            current_list[date_index][symbol] = (Decimal(0), Decimal(0), Decimal(0))
-        current_quantity, current_amount, current_fees = current_list[date_index][
-            symbol
-        ]
-        current_list[date_index][symbol] = (
-            current_quantity + quantity,
-            current_amount + amount,
-            current_fees + fees,
-        )
+        return self.tax_year_start_date <= date <= self.tax_year_end_date
 
     def add_acquisition(
         self,
-        portfolio: Dict[str, Decimal],
+        portfolio: dict[str, Decimal],
         acquisition_list: HmrcTransactionLog,
         transaction: BrokerTransaction,
     ) -> None:
+        """Add new acquisition to the given list."""
         symbol = transaction.symbol
         quantity = transaction.quantity
         if symbol is None:
@@ -128,7 +104,7 @@ class CapitalGainsCalculator:
             if transaction.amount != -calculated_amount:
                 raise CalculatedAmountDiscrepancy(transaction, -calculated_amount)
             amount = -transaction.amount
-        CapitalGainsCalculator.add_to_list(
+        add_to_list(
             acquisition_list,
             date_to_index(transaction.date),
             symbol,
@@ -139,10 +115,11 @@ class CapitalGainsCalculator:
 
     def add_disposal(
         self,
-        portfolio: Dict[str, Decimal],
+        portfolio: dict[str, Decimal],
         disposal_list: HmrcTransactionLog,
         transaction: BrokerTransaction,
     ) -> None:
+        """Add new disposal to the given list."""
         symbol = transaction.symbol
         quantity = transaction.quantity
         if symbol is None:
@@ -173,7 +150,7 @@ class CapitalGainsCalculator:
         )
         if amount != calculated_amount:
             raise CalculatedAmountDiscrepancy(transaction, calculated_amount)
-        CapitalGainsCalculator.add_to_list(
+        add_to_list(
             disposal_list,
             date_to_index(transaction.date),
             symbol,
@@ -182,21 +159,18 @@ class CapitalGainsCalculator:
             self.converter.to_gbp_for(transaction.fees, transaction),
         )
 
-    @staticmethod
-    def swift_date(date: datetime.date) -> str:
-        return date.strftime("%d/%m/%Y")
-
     def convert_to_hmrc_transactions(
         self,
-        transactions: List[BrokerTransaction],
-    ) -> Tuple[HmrcTransactionLog, HmrcTransactionLog]:
+        transactions: list[BrokerTransaction],
+    ) -> tuple[HmrcTransactionLog, HmrcTransactionLog]:
+        """Convert broker transactions to HMRC transactions."""
         # We keep a balance per broker,currency pair
-        balance: Dict[Tuple[str, str], Decimal] = defaultdict(lambda: Decimal(0))
+        balance: dict[tuple[str, str], Decimal] = defaultdict(lambda: Decimal(0))
         dividends = Decimal(0)
         dividends_tax = Decimal(0)
         interest = Decimal(0)
         total_sells = Decimal(0)
-        portfolio: Dict[str, Decimal] = {}
+        portfolio: dict[str, Decimal] = {}
         acquisition_list: HmrcTransactionLog = {}
         disposal_list: HmrcTransactionLog = {}
 
@@ -216,7 +190,6 @@ class CapitalGainsCalculator:
                     raise AmountMissingError(transaction)
                 new_balance += transaction.amount
                 self.add_disposal(portfolio, disposal_list, transaction)
-                # TODO: cleanup
                 if self.date_in_tax_year(transaction.date):
                     total_sells += self.converter.to_gbp_for(
                         transaction.amount, transaction
@@ -230,7 +203,7 @@ class CapitalGainsCalculator:
                 gbp_fees = self.converter.to_gbp_for(transaction.fees, transaction)
                 if transaction.symbol is None:
                     raise SymbolMissingError(transaction)
-                CapitalGainsCalculator.add_to_list(
+                add_to_list(
                     acquisition_list,
                     date_to_index(transaction.date),
                     transaction.symbol,
@@ -289,14 +262,15 @@ class CapitalGainsCalculator:
         print("")
         return acquisition_list, disposal_list
 
+    @staticmethod
     def process_acquisition(
-        self,
         acquisition_list: HmrcTransactionLog,
         bed_and_breakfast_list: HmrcTransactionLog,
-        portfolio: Dict[str, Tuple[Decimal, Decimal]],
+        portfolio: dict[str, tuple[Decimal, Decimal]],
         symbol: str,
         date_index: int,
-    ) -> List[CalculationEntry]:
+    ) -> list[CalculationEntry]:
+        """Process single acquisition."""
         acquisition_quantity, acquisition_amount, acquisition_fees = acquisition_list[
             date_index
         ][symbol]
@@ -311,18 +285,22 @@ class CapitalGainsCalculator:
         assert acquisition_amount > 0
         bed_and_breakfast_quantity = Decimal(0)
         bed_and_breakfast_amount = Decimal(0)
+        bed_and_breakfast_fees = Decimal(0)
         if acquisition_quantity > 0:
             acquisition_price = acquisition_amount / acquisition_quantity
             if has_key(bed_and_breakfast_list, date_index, symbol):
                 (
                     bed_and_breakfast_quantity,
                     bed_and_breakfast_amount,
-                    bed_and_breakfast_fees,
+                    _bed_and_breakfast_fees,
                 ) = bed_and_breakfast_list[date_index][symbol]
                 assert bed_and_breakfast_quantity <= acquisition_quantity
                 acquisition_amount -= bed_and_breakfast_quantity * acquisition_price
                 acquisition_amount += bed_and_breakfast_amount
                 assert acquisition_amount > 0
+                bed_and_breakfast_fees = (
+                    acquisition_fees * bed_and_breakfast_quantity / acquisition_quantity
+                )
                 calculation_entries.append(
                     CalculationEntry(
                         rule_type=RuleType.BED_AND_BREAKFAST,
@@ -349,24 +327,26 @@ class CapitalGainsCalculator:
                     amount=-(acquisition_amount - bed_and_breakfast_amount),
                     new_quantity=current_quantity + acquisition_quantity,
                     new_pool_cost=current_amount + acquisition_amount,
-                    fees=acquisition_fees,
+                    fees=acquisition_fees - bed_and_breakfast_fees,
                     allowable_cost=original_acquisition_amount,
                 )
             )
         return calculation_entries
 
+    @staticmethod
     def process_disposal(
-        self,
         acquisition_list: HmrcTransactionLog,
         disposal_list: HmrcTransactionLog,
         bed_and_breakfast_list: HmrcTransactionLog,
-        portfolio: Dict[str, Tuple[Decimal, Decimal]],
+        portfolio: dict[str, tuple[Decimal, Decimal]],
         symbol: str,
         date_index: int,
-    ) -> Tuple[Decimal, List[CalculationEntry]]:
+    ) -> tuple[Decimal, list[CalculationEntry]]:
+        """Process single disposal."""
         disposal_quantity, proceeds_amount, disposal_fees = disposal_list[date_index][
             symbol
         ]
+        original_disposal_quantity = disposal_quantity
         disposal_price = proceeds_amount / disposal_quantity
         current_quantity, current_amount = portfolio[symbol]
         assert disposal_quantity <= current_quantity
@@ -374,33 +354,24 @@ class CapitalGainsCalculator:
         calculation_entries = []
         # Same day rule is first
         if has_key(acquisition_list, date_index, symbol):
-            same_day_quantity, same_day_amount, same_day_fees = acquisition_list[
+            same_day_quantity, same_day_amount, _same_day_fees = acquisition_list[
                 date_index
             ][symbol]
-            bed_and_breakfast_quantity = Decimal(0)
-            if has_key(bed_and_breakfast_list, date_index, symbol):
-                (
-                    bed_and_breakfast_quantity,
-                    _bb_amount,
-                    _bb_fees,
-                ) = bed_and_breakfast_list[date_index][symbol]
-            assert bed_and_breakfast_quantity <= same_day_quantity
-            available_quantity = min(
-                disposal_quantity, same_day_quantity - bed_and_breakfast_quantity
-            )
+            available_quantity = min(disposal_quantity, same_day_quantity)
             if available_quantity > 0:
                 acquisition_price = same_day_amount / same_day_quantity
                 same_day_proceeds = available_quantity * disposal_price
                 same_day_allowable_cost = available_quantity * acquisition_price
                 same_day_gain = same_day_proceeds - same_day_allowable_cost
                 chargeable_gain += same_day_gain
-                # print(
-                #     f"SAME DAY"
-                #     f", quantity {available_quantity}"
-                #     f", gain {same_day_gain}
-                #     f", disposal price {disposal_price}"
-                #     f", acquisition price {acquisition_price}"
-                # )
+                LOGGER.debug(
+                    "SAME DAY, quantity %d, gain %s, disposal price %s, "
+                    "acquisition price %s",
+                    available_quantity,
+                    same_day_gain,
+                    disposal_price,
+                    acquisition_price,
+                )
                 disposal_quantity -= available_quantity
                 proceeds_amount -= available_quantity * disposal_price
                 current_quantity -= available_quantity
@@ -408,6 +379,7 @@ class CapitalGainsCalculator:
                 current_amount -= available_quantity * acquisition_price
                 if current_quantity == 0:
                     assert current_amount == 0, f"current amount {current_amount}"
+                fees = disposal_fees * available_quantity / original_disposal_quantity
                 calculation_entries.append(
                     CalculationEntry(
                         rule_type=RuleType.SAME_DAY,
@@ -415,7 +387,7 @@ class CapitalGainsCalculator:
                         amount=same_day_proceeds,
                         gain=same_day_gain,
                         allowable_cost=same_day_allowable_cost,
-                        fees=same_day_fees,
+                        fees=fees,
                         new_quantity=current_quantity,
                         new_pool_cost=current_amount,
                     )
@@ -423,14 +395,15 @@ class CapitalGainsCalculator:
 
         # Bed and breakfast rule next
         if disposal_quantity > 0:
-            for i in range(30):
+            for i in range(BED_AND_BREAKFAST_DAYS):
                 search_index = date_index + i + 1
                 if has_key(acquisition_list, search_index, symbol):
                     (
                         acquisition_quantity,
                         acquisition_amount,
-                        acquisition_fees,
+                        _acquisition_fees,
                     ) = acquisition_list[search_index][symbol]
+
                     bed_and_breakfast_quantity = Decimal(0)
                     if has_key(bed_and_breakfast_list, search_index, symbol):
                         (
@@ -439,9 +412,24 @@ class CapitalGainsCalculator:
                             _bb_fees,
                         ) = bed_and_breakfast_list[search_index][symbol]
                     assert bed_and_breakfast_quantity <= acquisition_quantity
+
+                    same_day_quantity = Decimal(0)
+                    if has_key(disposal_list, search_index, symbol):
+                        (
+                            same_day_quantity,
+                            _same_day_amount,
+                            _same_day_fees,
+                        ) = disposal_list[search_index][symbol]
+                    assert same_day_quantity <= acquisition_quantity
+
                     # This can be some management fee entry or already used
                     # by bed and breakfast rule
-                    if acquisition_quantity - bed_and_breakfast_quantity == 0:
+                    if (
+                        acquisition_quantity
+                        - same_day_quantity
+                        - bed_and_breakfast_quantity
+                        == 0
+                    ):
                         continue
                     print(
                         f"WARNING: Bed and breakfasting for {symbol}."
@@ -450,7 +438,9 @@ class CapitalGainsCalculator:
                     )
                     available_quantity = min(
                         disposal_quantity,
-                        acquisition_quantity - bed_and_breakfast_quantity,
+                        acquisition_quantity
+                        - same_day_quantity
+                        - bed_and_breakfast_quantity,
                     )
                     acquisition_price = acquisition_amount / acquisition_quantity
                     bed_and_breakfast_proceeds = available_quantity * disposal_price
@@ -461,13 +451,14 @@ class CapitalGainsCalculator:
                         bed_and_breakfast_proceeds - bed_and_breakfast_allowable_cost
                     )
                     chargeable_gain += bed_and_breakfast_gain
-                    # print(
-                    #     f"BED & BREAKFAST"
-                    #     f", quantity {available_quantity}"
-                    #     f", gain {bed_and_breakfast_gain}"
-                    #     f", disposal price {disposal_price}"
-                    #     f", acquisition price {acquisition_price}"
-                    # )
+                    LOGGER.debug(
+                        "BED & BREAKFAST, quantity %d, gain %s, disposal price %s, "
+                        "acquisition price %s",
+                        available_quantity,
+                        bed_and_breakfast_gain,
+                        disposal_price,
+                        acquisition_price,
+                    )
                     disposal_quantity -= acquisition_quantity
                     proceeds_amount -= available_quantity * disposal_price
                     current_price = current_amount / current_quantity
@@ -476,7 +467,7 @@ class CapitalGainsCalculator:
                     current_amount -= amount_delta
                     if current_quantity == 0:
                         assert current_amount == 0, f"current amount {current_amount}"
-                    CapitalGainsCalculator.add_to_list(
+                    add_to_list(
                         bed_and_breakfast_list,
                         search_index,
                         symbol,
@@ -484,6 +475,9 @@ class CapitalGainsCalculator:
                         amount_delta,
                         Decimal(0),
                     )
+                    fees = (
+                        disposal_fees * available_quantity / original_disposal_quantity
+                    )
                     calculation_entries.append(
                         CalculationEntry(
                             rule_type=RuleType.BED_AND_BREAKFAST,
@@ -491,8 +485,7 @@ class CapitalGainsCalculator:
                             amount=bed_and_breakfast_proceeds,
                             gain=bed_and_breakfast_gain,
                             allowable_cost=bed_and_breakfast_allowable_cost,
-                            # TODO: support fees
-                            fees=acquisition_fees,
+                            fees=fees,
                             bed_and_breakfast_date_index=search_index,
                             new_quantity=current_quantity,
                             new_pool_cost=current_amount,
@@ -503,19 +496,21 @@ class CapitalGainsCalculator:
                 current_amount * Decimal(disposal_quantity) / Decimal(current_quantity)
             )
             chargeable_gain += proceeds_amount - allowable_cost
-            # print(
-            #     f"SECTION 104"
-            #     f", quantity {disposal_quantity}"
-            #     f", gain {proceeds_amount - allowable_cost}"
-            #     f", proceeds amount {proceeds_amount}"
-            #     f", allowable cost {allowable_cost}"
-            # )
+            LOGGER.debug(
+                "SECTION 104, quantity %d, gain %s, proceeds amount %s, "
+                "allowable cost %s",
+                disposal_quantity,
+                proceeds_amount - allowable_cost,
+                proceeds_amount,
+                allowable_cost,
+            )
             current_quantity -= disposal_quantity
             current_amount -= allowable_cost
             if current_quantity == 0:
                 assert (
                     round_decimal(current_amount, 10) == 0
                 ), f"current amount {current_amount}"
+            fees = disposal_fees * disposal_quantity / original_disposal_quantity
             calculation_entries.append(
                 CalculationEntry(
                     rule_type=RuleType.SECTION_104,
@@ -523,8 +518,7 @@ class CapitalGainsCalculator:
                     amount=proceeds_amount,
                     gain=proceeds_amount - allowable_cost,
                     allowable_cost=allowable_cost,
-                    # CHECK THIS!
-                    fees=disposal_fees,
+                    fees=fees,
                     new_quantity=current_quantity,
                     new_pool_cost=current_amount,
                 )
@@ -538,7 +532,8 @@ class CapitalGainsCalculator:
         acquisition_list: HmrcTransactionLog,
         disposal_list: HmrcTransactionLog,
     ) -> CapitalGainsReport:
-        begin_index = date_to_index(internal_start_date)
+        """Calculate capital gain and return generated report."""
+        begin_index = date_to_index(INTERNAL_START_DATE)
         tax_year_start_index = date_to_index(self.tax_year_start_date)
         end_index = date_to_index(self.tax_year_end_date)
         disposal_count = 0
@@ -547,7 +542,7 @@ class CapitalGainsCalculator:
         capital_gain = Decimal(0)
         capital_loss = Decimal(0)
         bed_and_breakfast_list: HmrcTransactionLog = {}
-        portfolio: Dict[str, Tuple[Decimal, Decimal]] = {}
+        portfolio: dict[str, tuple[Decimal, Decimal]] = {}
         calculation_log: CalculationLog = {}
         for date_index in range(begin_index, end_index + 1):
             if date_index in acquisition_list:
@@ -588,11 +583,13 @@ class CapitalGainsCalculator:
                             transaction_disposal_proceeds - transaction_capital_gain
                         )
                         transaction_quantity = disposal_list[date_index][symbol][0]
-                        # print(
-                        #     f"DISPOSAL on {date_from_index(date_index)} of {symbol}"
-                        #     f", quantity {transaction_quantity}: "
-                        #     f"capital gain: ${round_decimal(transaction_capital_gain, 2)}"
-                        # )
+                        LOGGER.debug(
+                            "DISPOSAL on %s of %s, quantity %d, capital gain $%s",
+                            date_from_index(date_index),
+                            symbol,
+                            transaction_quantity,
+                            round_decimal(transaction_capital_gain, 2),
+                        )
                         calculated_quantity = Decimal(0)
                         calculated_proceeds = Decimal(0)
                         calculated_gain = Decimal(0)
@@ -619,7 +616,7 @@ class CapitalGainsCalculator:
                         else:
                             capital_loss += transaction_capital_gain
         print("\nSecond pass completed")
-        allowance = self.CAPITAL_GAIN_ALLOWANCES.get(self.tax_year)
+        allowance = CAPITAL_GAIN_ALLOWANCES.get(self.tax_year)
         return CapitalGainsReport(
             self.tax_year,
             portfolio,
@@ -634,16 +631,28 @@ class CapitalGainsCalculator:
 
 
 def main() -> int:
+    """Run main function."""
     # Throw exception on accidental float usage
     decimal.getcontext().traps[decimal.FloatOperation] = True
     args = create_parser().parse_args()
 
+    if args.version:
+        print(f"cgt-calc {importlib.metadata.version(__package__)}")
+        return 0
+
+    if args.report == "":
+        print("error: report name can't be empty")
+        return 1
+
+    default_logging_level = logging.DEBUG if args.verbose else logging.WARNING
+    logging.basicConfig(level=default_logging_level)
+
     # Read data from input files
     broker_transactions = read_broker_transactions(args.schwab, args.trading212)
     converter = CurrencyConverter(read_gbp_prices_history(args.gbp_history))
     initial_prices = InitialPrices(read_initial_prices(args.initial_prices))
 
-    calculator = CapitalGainsCalculator(args.tax_year, converter, initial_prices)
+    calculator = CapitalGainsCalculator(args.year, converter, initial_prices)
     # First pass converts broker transactions to HMRC transactions.
     # This means applying same day rule and collapsing all transactions with
     # same type in the same day.
@@ -655,17 +664,19 @@ def main() -> int:
     # Second pass calculates capital gain tax for the given tax year
     report = calculator.calculate_capital_gain(acquisition_list, disposal_list)
     print(report)
-    if args.report:
+
+    # Generate PDF report
+    if not args.no_report:
         render_latex.render_calculations(
-            report.calculation_log,
-            tax_year=report.tax_year,
-            date_from_index=date_from_index,
-            output_file=args.report,
+            report,
+            output_path=Path(args.report),
+            skip_pdflatex=args.no_pdflatex,
         )
     print("All done!")
 
     return 0
 
 
-if __name__ == "__main__":
+def init() -> None:
+    """Entry point."""
     sys.exit(main())
